#![allow(dead_code)]
use serde::{Serialize, Deserialize};
    
struct Context {
    client: reqwest::Client,
    api_key: Option<String>,
    url: String,

    // cache for login()
    mfa_key: Option<String>,
    device: Option<String>,
}

#[derive(Deserialize)]
struct ErrorResponse {
    error: String,
}

#[derive(Serialize)]
struct LoginRequest<'a> {
    email: &'a str,
    password: &'a str,
    device: &'a str,
}

#[derive(Deserialize)]
struct LoginResponse {
    name: String,
    email: String,
    mfa_enabled: bool,
    mfa_key: Option<String>,
    api_key: Option<String>,
}

pub enum Error {
    FIDO(String),
    Json(String),
    Auth(String),
    Http(String),
    Internal(String)
}

pub const DEFAULT_URL: &str = "https://app.simplelogin.io/";

impl Context {
    pub fn new(api_key: Option<&str>) -> Self {
        Self {
            client: reqwest::Client::new(),
            api_key: api_key.map(|key| key.to_owned()),
            url: DEFAULT_URL.to_owned(),
            mfa_key: None,
            device: None,
        }
    }

    pub fn new_with_url(api_key: Option<&str>, url: &str) -> Self {
        Self {
            client: reqwest::Client::new(),
            api_key: api_key.map(|key| key.to_owned()),
            url: url.to_owned(),
            mfa_key: None,
            device: None
        }
    }

    fn parse_error_json(error_body: &String) -> String {
        match serde_json::from_str(error_body) {
            Ok::<ErrorResponse, _>(obj) => obj.error,
            Err(_) => "Could not parse error response JSON".to_owned(),
        }
    }

    async fn perform_request(
        req_obj: reqwest::RequestBuilder
    ) -> Result<String, Error> {
        match req_obj.send().await {
            Ok(resp) => {
                let status = resp.status();
                let resp_text = resp.text().await
                    .map_err(|e| Error::Http(e.to_string()))?;
                return if status.is_success() {
                    Ok(resp_text)
                } else if status.is_client_error() {
                    Err(Error::Auth(Self::parse_error_json(&resp_text)))
                } else {
                    Err(Error::Http("Unknown error! Response code: ".to_owned()
                        + status.as_str()))
                }
            },
            Err(err) => {
                Err(Error::Http(err.to_string()))
            },
        }
    }

    pub async fn login(
        &mut self,
        email: &str,
        password: &str,
        device: &str
    ) -> Result<bool, Error> {
        let req_body = LoginRequest {email, password, device};
        let response = Self::perform_request(self.client.post(self.url.to_owned())
            .header("Content-Type", "application/json")
            .body(serde_json::to_string(&req_body).unwrap()));
        let body = response.await?;
        let resp_obj: LoginResponse = serde_json::from_str(body.as_str())
            .map_err(|_| Error::Json("Could not parse response json"
                .to_owned()))?;
        if resp_obj.mfa_enabled {
            if resp_obj.mfa_key.is_some() {
                self.device = Some(device.to_owned());
                self.mfa_key = resp_obj.mfa_key;
                Ok(true)
            } else {
                Err(Error::Json("Error response did not contain mfa key"
                    .to_owned()))
            }
        } else {
            if resp_obj.api_key.is_some() {
                self.api_key = resp_obj.api_key;
                self.mfa_key = None;
                self.device = None;
                Ok(true)
            } else {
                Err(Error::Json("Response did not contain api key".to_owned()))
            }
        }
    }

    pub async fn mfa(&mut self, mfa_token: &str) -> Result<(), Error> {
        // TODO implement
        Ok(())
    }
}
